using System;
using System.Linq;
using UnityEngine;
using System.Collections;

public enum StatModifierType
{
    Additive
}

public enum StatClampType
{
    RawFloat,
    Round,
    Floor
}

[System.Serializable]
public class StatModifier : ICloneable
{
    public string StatName;
    public StatModifierType ModifierType;
    public float Amount;
    public float Duration;    

    public StatModifier(string statName, StatModifierType statModifierType, float amount, float duration)
    {
        StatName = statName;
        ModifierType = statModifierType;
        Amount = amount;
        Duration = duration;
    }

    public object Clone()
    {
        return MemberwiseClone();
    }

    public override bool Equals(object obj)
    {
        StatModifier other = obj as StatModifier;
        if (other != null && other.StatName == StatName && other.ModifierType == ModifierType && other.Amount == Amount && other.Duration == Duration)
        {
            return true;
        }

        return false;
    }

    public override int GetHashCode()
    {
        return base.GetHashCode();
    }
}

public class StatController : MonoBehaviour
{
    [System.Serializable]
    public class StatEntry
    {
        public string Name;
        public float StartingValue;        
        public float MinValue;
        public float MaxValue;
        public StatClampType ClampType;
    }

    public StatEntry[] Stats;
    private ArrayList modifiers = new ArrayList();

	void Start()
	{
	    ValidateAllStats();
	}
	
	void Update()
    {
	    for (int i = modifiers.Count - 1; i >= 0; i--)
	    {
	        StatModifier statMod = (StatModifier)modifiers[i];
            if (statMod.Duration > 0.0f)
            {
                statMod.Duration -= Time.deltaTime;
                if (statMod.Duration < 0.0f)
                {
                    modifiers.RemoveAt(i);
                    i--;
                }
            }
	    }
	}

    void OnGUI()
    {        
        GUI.Label(new Rect(5.0f, 25.0f, 250.0f, 25.0f), "--- Stats ---");
        float yValue = 50.0f;
        foreach (StatEntry statEntry in Stats)
        {
            GUI.Label(new Rect(5.0f, yValue, 250.0f, 25.0f), statEntry.Name + ": " + GetStatValue(statEntry.Name));
            yValue += 25.0f;
        }

        GUI.Label(new Rect(5.0f, yValue, 250.0f, 25.0f), "--- Active Modifiers ---");
        yValue += 25.0f;

        foreach (StatModifier statModifier in modifiers)
        {
            string modifierAmount;
            if (statModifier.Amount >= 0.0f)
            {
                modifierAmount = "+" + statModifier.Amount;
            }
            else
            {
                modifierAmount = statModifier.Amount.ToString();
            }

            if (statModifier.Duration > 0.0f)
            {
                GUI.Label(new Rect(5.0f, yValue, 250.0f, 25.0f), statModifier.StatName + ": " + modifierAmount + " (" + statModifier.Duration + ")");
            }
            else
            {
                GUI.Label(new Rect(5.0f, yValue, 250.0f, 25.0f), statModifier.StatName + ": " + modifierAmount);
            }
            
            yValue += 25.0f;   
        }        
    }

    private StatEntry GetStatEntry(string statName)
    {
        return Stats.FirstOrDefault(statInfo => statInfo.Name == statName);
    }

    public float GetStatValue(string statName)
    {
        StatEntry statEntry = GetStatEntry(statName);
        if (statEntry == null)
        {
            Debug.LogWarning("StatController::GetStatValue() - Tried to get a stat value for a stat that doesn't exist: " + statName);
            return -1.0f;
        }

        float returnValue = statEntry.StartingValue;
        returnValue = ApplyModifiersToStatValue(statName, returnValue);
        returnValue = Mathf.Clamp(returnValue, statEntry.MinValue, statEntry.MaxValue);
        returnValue = ClampStatValue(statEntry.ClampType, returnValue);
        return returnValue;
    }

    private float ClampStatValue(StatClampType clampType, float baseValue)
    {
        switch (clampType)
        {
            case StatClampType.RawFloat:
                return baseValue;
            case StatClampType.Floor:
                return Mathf.Floor(baseValue);
            case StatClampType.Round:
                return Mathf.Round(baseValue);
        }

        return baseValue;
    }

    private float ApplyModifiersToStatValue(string statName, float baseValue)
    {
        foreach (StatModifier statModifier in modifiers)
        {
            if (statModifier.StatName == statName)
            {
                if (statModifier.ModifierType == StatModifierType.Additive)
                {
                    baseValue += statModifier.Amount;
                }
                else
                {
                    Debug.LogWarning("StatController::ApplyModifiersToStatValue() - Unsupported stat modifier type!");
                }
            }
        }

        return baseValue;
    }

    public void RemoveModifier(StatModifier statModifier)
    {
        modifiers.Remove(statModifier);
    }

    public void AddModifier(StatModifier statModifier)
    {
        bool foundStat = Stats.Any(statEntry => statEntry.Name == statModifier.StatName);
        if (foundStat == false)
        {
            Debug.LogError("StatController::AddModifier() - Trying to add a modifier for a stat that doesn't exist! Stat: " + statModifier.StatName);
            return;
        }

        modifiers.Add(statModifier);
    }

    private void ValidateAllStats()
    {
        ArrayList usedNames = new ArrayList();
        foreach (StatEntry statInfo in Stats)
        {
            if (statInfo.Name.Length == 0)
            {
                Debug.LogError("StatController::ValidateAllStats() - Stat shouldn't have a name with a length of 0!");
            }

            if (usedNames.Contains(statInfo.Name))
            {
                Debug.LogError("StatController::ValidateAllStats() - Stat [" + statInfo.Name + "] is in the Stats list more than once!");
            }

            usedNames.Add(statInfo.Name);

            if (statInfo.MinValue > statInfo.MaxValue)
            {
                Debug.LogError("StatController::ValidateAllStats() - Stat [" + statInfo.Name + "] has a MinValue > MaxValue!");
            }

            if (statInfo.StartingValue < statInfo.MinValue || statInfo.StartingValue > statInfo.MaxValue)
            {
                Debug.LogError("StatController::ValidateAllStats() - Stat [" + statInfo.Name + "] starting value of " + statInfo.StartingValue + " is not within the MinValue/MaxValue range of " + statInfo.MinValue + " -> " + statInfo.MaxValue);
            }
        }
    }
}
